# nsolid-registry-etcd

This is the N|Solid Registry API layer for talking to an Etcd cluster.

It contains code for both the client (registration) and the Proxy (list/lookup)

Client Pseudocode
```js
var registry = require("nsolid-registry-etcd")("localhost:4001")
var agent = require("nsolid_agent")
var appData = {app: "foobar", id: agent.id, address: agent.socket}

var ttl = 10 // seconds
var pause = 5000 // ms (5 seconds)

// Set up a repeated registration -- ttl will expire record so must keep
// bumping the ttl to stay listed as "live"
function register() {
  registry.register(appData, ttl, function cb(err, reply) {
    if (err) {
      // handle error
      return
    }
    var to = setTimeout(register, pause)
    to.unref()
  })
}
```

Proxy Pseudocode
```js
var registry = require("nsolid-registry-etcd")("localhost:4001")

function listRoute(req, res) {
  // hand-wavy req parsing
  var app = req.params.app
  registry.list(app, function (err, hosts) {
    if (err) {
      // handle error
      res.statusCode = 503
      return res.end(err)
    }
    res.end(JSON.stringify(hosts))
  })
}

function pingRoute(req, res) {
  // hand-wavy req parsing
  var app = req.params.app
  var id = req.params.id
  if (id == null) {
    // ping is whitelisted so we can broadcast
    registry.list(app, function (err, hosts) {
      if (err) {
        // error handling
        return
      }
      broadcast("ping", null, hosts).pipe(res)
    })
    return
  }
  registry.lookup(app, id, function (err, hosts) {
    if (err) {
      // handle error
      res.statusCode = 503
      return res.end(err)
    }
    singleProxy("ping", null, host).pipe(res)
  })
}

function broadcast(command, data, hosts) {
  // hypothetical function that does a fan-out of the command/data to each host
  // and streams back the replies as \n delimited JSON
}

function singleProxy(command, data, host) {
  // hypothetical function that sends command/data to a single host and streams
  // back the JSON string
}
```

## Usage

### `var registry = new Registry(etcdClusterHost, etcdOptions)`

Instantiate the registry object. Requires an ip/port of an Etcd server on the cluster.

The `etcdCluster` parameter can specify multiple ip/port values, separated
by commas - "','".  Spaces before/after each entry are removed.

Each ip/port value can be prefixed with `etcd://`, which will simply be ignored.

Examples:

    new Registry("4001")
    new Registry("localhost:4001")
    new Registry("localhost:4001, etcd://otherhost:4001")
    new Registry("4001, 4002, 4003")
    new Registry("4001, 4002, etcd://4003")

The `etcdOptions` parameter is an options object passed to the etcdjs package's
[`etcd()` API](https://github.com/mafintosh/etcdjs#store--etcdhost-opts).  The
`json` property will always be passed as `true`, but you may want to set an
option like `timeout`.  That value is the number of milliseconds to wait for
a response from an etcd server.

Can also be called as a factory method.

### `registry.register(appData[, ttl][, callback])`

Register by adding a record with a TTL to Etcd. Default TTL is 60 seconds, and the process is expected to do the re-registration periodically on its own. This is NOT handled for you.

The `appData` object requries at least these three keys:
* app -- the classification of this application, e.g. "prod-webserver"
* id -- the unique run-instance of this execution, should be require("nsolid_agent").id
* address -- the socket that the NSolid agent is listening on. IP + port

Callback will be called with `callback(err, reply)` the error and reply are both directly populated from Etcd

### `registry.list(appData, callback)`

List the known processes running with the same "application" type.

Callback will be called with `callback(err, processes)` where `processes` is an Array containing the `appData` sent by each process when registering.

### `registry.lookup(appData, id, callback)`

Same as list but for a specific process. Requires the `id` as well and will return only the `appData` that process registered with.


## Running Etcd

For testing and development purposes it is easiest to run Etcd in single-machine mode locally.

1. Download Etcd from here: https://github.com/coreos/etcd/releases -- I'm currently testing with 2.0.10
2. untar & run like so: `./etcd -data-dir machine0 -name machine0`

(The dir and name can be changed to whatever you prefer)

The tests of this repository expect Etcd to be running locally.

## Etcd library selection

Right now I'm using http://npm.im/etcdjs for a couple reasons:
1. I know and trust the author
2. fewer dependencies
3. the most popular one http://npm.im/node-etcd requires `deasync` which is an anti-pattern
4. `node-etcd` is also written in coffee which makes it harder for me to review and patch

The most important thing that `node-etcd` does that this one is missing is the ability to specify a pool of Etcd servers to connect to with some reconnection logic.

This is something we'll need to consider in the near term to help with resilience of talking to the N|Solid agents.

Failure scenarios:
1. Process needs to register during a leader election & can't write
  * may (for unknown period of time) not be apparent the process is online
2. Etcd node that the process was talking to went down but cluster lives on
  * process will be unable to renew registration lease
  * process will appear down to N|Solid proxies talking to the live part of the Etcd cluster
3. Others?
